MODULE agrif_oce_interp
   !!======================================================================
   !!                   ***  MODULE  agrif_oce_interp  ***
   !! AGRIF: interpolation package for the ocean dynamics (OPA)
   !!======================================================================
   !! History :  2.0  !  2002-06  (L. Debreu)  Original cade
   !!            3.2  !  2009-04  (R. Benshila) 
   !!            3.6  !  2014-09  (R. Benshila) 
   !!----------------------------------------------------------------------
#if defined key_agrif
   !!----------------------------------------------------------------------
   !!   'key_agrif'                                              AGRIF zoom
   !!----------------------------------------------------------------------
   !!   Agrif_tra     :
   !!   Agrif_dyn     : 
   !!   Agrif_ssh     :
   !!   Agrif_dyn_ts  :
   !!   Agrif_dta_ts  :
   !!   Agrif_ssh_ts  :
   !!   Agrif_avm     : 
   !!   interpu       :
   !!   interpv       :
   !!----------------------------------------------------------------------
   USE par_oce
   USE oce
   USE dom_oce      
   USE zdf_oce
   USE agrif_oce
   USE phycst
   USE dynspg_ts, ONLY: un_adv, vn_adv
   !
   USE in_out_manager
   USE agrif_oce_sponge
   USE lib_mpp
 
   IMPLICIT NONE
   PRIVATE

   PUBLIC   Agrif_dyn, Agrif_ssh, Agrif_dyn_ts, Agrif_dyn_ts_flux, Agrif_ssh_ts, Agrif_dta_ts
   PUBLIC   Agrif_tra, Agrif_avm
   PUBLIC   interpun , interpvn
   PUBLIC   interptsn, interpsshn, interpavm
   PUBLIC   interpunb, interpvnb , interpub2b, interpvb2b
   PUBLIC   interpe3t, interpumsk, interpvmsk

#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/NST 4.0 , NEMO Consortium (2018)
   !! $Id$
   !! Software governed by the CeCILL license (see ./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS

   SUBROUTINE Agrif_tra
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_tra  ***
      !!----------------------------------------------------------------------
      !
      IF( Agrif_Root() )   RETURN
      !
      Agrif_SpecialValue    = 0._wp
      Agrif_UseSpecialValue = .TRUE.
      !
      CALL Agrif_Bc_variable( tsn_id, procname=interptsn )
      !
      Agrif_UseSpecialValue = .FALSE.
      !
   END SUBROUTINE Agrif_tra


   SUBROUTINE Agrif_dyn( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_DYN  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   kt
      !
      INTEGER ::   ji, jj, jk       ! dummy loop indices
      INTEGER ::   ibdy1, jbdy1, ibdy2, jbdy2
      REAL(wp), DIMENSION(jpi,jpj) ::   zub, zvb
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      Agrif_SpecialValue    = 0._wp
      Agrif_UseSpecialValue = ln_spc_dyn
      !
      CALL Agrif_Bc_variable( un_interp_id, procname=interpun )
      CALL Agrif_Bc_variable( vn_interp_id, procname=interpvn )
      !
      Agrif_UseSpecialValue = .FALSE.
      !
      ! --- West --- !
      ibdy1 = 2
      ibdy2 = 1+nbghostcells 
      !
      IF( .NOT.ln_dynspg_ts ) THEN  ! Store transport
         DO ji = mi0(ibdy1), mi1(ibdy2)
            ua_b(ji,:) = 0._wp

            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ua_b(ji,jj) = ua_b(ji,jj) + e3u_a(ji,jj,jk) * ua(ji,jj,jk) * umask(ji,jj,jk)
               END DO
            END DO

            DO jj = 1, jpj
               ua_b(ji,jj) = ua_b(ji,jj) * r1_hu_a(ji,jj)
            END DO
         END DO
      ENDIF
      !
      DO ji = mi0(ibdy1), mi1(ibdy2)
         zub(ji,:) = 0._wp    ! Correct transport
         DO jk = 1, jpkm1
            DO jj = 1, jpj
               zub(ji,jj) = zub(ji,jj) & 
                  & + e3u_a(ji,jj,jk)  * ua(ji,jj,jk)*umask(ji,jj,jk)
            END DO
         END DO
         DO jj=1,jpj
            zub(ji,jj) = zub(ji,jj) * r1_hu_a(ji,jj)
         END DO
            
         DO jk = 1, jpkm1
            DO jj = 1, jpj
               ua(ji,jj,jk) = ( ua(ji,jj,jk) + ua_b(ji,jj)-zub(ji,jj)) * umask(ji,jj,jk)
            END DO
         END DO
      END DO
            
      IF( ln_dynspg_ts ) THEN       ! Set tangential velocities to time splitting estimate
         DO ji = mi0(ibdy1), mi1(ibdy2)
            zvb(ji,:) = 0._wp
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  zvb(ji,jj) = zvb(ji,jj) + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
               END DO
            END DO
            DO jj = 1, jpj
               zvb(ji,jj) = zvb(ji,jj) * r1_hv_a(ji,jj)
            END DO
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  va(ji,jj,jk) = ( va(ji,jj,jk) + va_b(ji,jj)-zvb(ji,jj))*vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF

      ! --- East --- !
      ibdy1 = jpiglo-1-nbghostcells
      ibdy2 = jpiglo-2 
      !
      IF( .NOT.ln_dynspg_ts ) THEN  ! Store transport
         DO ji = mi0(ibdy1), mi1(ibdy2)
            ua_b(ji,:) = 0._wp
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  ua_b(ji,jj) = ua_b(ji,jj) & 
                      & + e3u_a(ji,jj,jk) * ua(ji,jj,jk) * umask(ji,jj,jk)
               END DO
            END DO
            DO jj = 1, jpj
               ua_b(ji,jj) = ua_b(ji,jj) * r1_hu_a(ji,jj)
            END DO
         END DO
      ENDIF
      !
      DO ji = mi0(ibdy1), mi1(ibdy2)
         zub(ji,:) = 0._wp    ! Correct transport
         DO jk = 1, jpkm1
            DO jj = 1, jpj
               zub(ji,jj) = zub(ji,jj) & 
                  & + e3u_a(ji,jj,jk)  * ua(ji,jj,jk) * umask(ji,jj,jk)
            END DO
         END DO
         DO jj=1,jpj
            zub(ji,jj) = zub(ji,jj) * r1_hu_a(ji,jj)
         END DO
            
         DO jk = 1, jpkm1
            DO jj = 1, jpj
               ua(ji,jj,jk) = ( ua(ji,jj,jk) & 
                 & + ua_b(ji,jj)-zub(ji,jj))*umask(ji,jj,jk)
            END DO
         END DO
      END DO
            
      IF( ln_dynspg_ts ) THEN       ! Set tangential velocities to time splitting estimate
         ibdy1 = jpiglo-nbghostcells
         ibdy2 = jpiglo-1 
         DO ji = mi0(ibdy1), mi1(ibdy2)
            zvb(ji,:) = 0._wp
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  zvb(ji,jj) = zvb(ji,jj) &
                     & + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
               END DO
            END DO
            DO jj = 1, jpj
               zvb(ji,jj) = zvb(ji,jj) * r1_hv_a(ji,jj)
            END DO
            DO jk = 1, jpkm1
               DO jj = 1, jpj
                  va(ji,jj,jk) = ( va(ji,jj,jk) & 
                      & + va_b(ji,jj)-zvb(ji,jj)) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF

      ! --- South --- !
      jbdy1 = 2
      jbdy2 = 1+nbghostcells 
      !
      IF( .NOT.ln_dynspg_ts ) THEN  ! Store transport
         DO jj = mj0(jbdy1), mj1(jbdy2)
            va_b(:,jj) = 0._wp
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  va_b(ji,jj) = va_b(ji,jj) & 
                      & + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
               END DO
            END DO
            DO ji=1,jpi
               va_b(ji,jj) = va_b(ji,jj) * r1_hv_a(ji,jj)     
            END DO
         END DO
      ENDIF
      !
      DO jj = mj0(jbdy1), mj1(jbdy2)
         zvb(:,jj) = 0._wp    ! Correct transport
         DO jk=1,jpkm1
            DO ji=1,jpi
               zvb(ji,jj) = zvb(ji,jj) & 
                  & + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
            END DO
         END DO
         DO ji = 1, jpi
            zvb(ji,jj) = zvb(ji,jj) * r1_hv_a(ji,jj)
         END DO

         DO jk = 1, jpkm1
            DO ji = 1, jpi
               va(ji,jj,jk) = ( va(ji,jj,jk) & 
                 & + va_b(ji,jj) - zvb(ji,jj) ) * vmask(ji,jj,jk)
            END DO
         END DO
      END DO
            
      IF( ln_dynspg_ts ) THEN       ! Set tangential velocities to time splitting estimate
         DO jj = mj0(jbdy1), mj1(jbdy2)
            zub(:,jj) = 0._wp
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  zub(ji,jj) = zub(ji,jj) & 
                     & + e3u_a(ji,jj,jk) * ua(ji,jj,jk) * umask(ji,jj,jk)
               END DO
            END DO
            DO ji = 1, jpi
               zub(ji,jj) = zub(ji,jj) * r1_hu_a(ji,jj)
            END DO
               
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ua(ji,jj,jk) = ( ua(ji,jj,jk) & 
                    & + ua_b(ji,jj) - zub(ji,jj) ) * umask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF

      ! --- North --- !
      jbdy1 = nlcj-1-nbghostcells
      jbdy2 = nlcj-2 
      !
      IF( .NOT.ln_dynspg_ts ) THEN  ! Store transport
         DO jj = mj0(jbdy1), mj1(jbdy2)
            va_b(:,jj) = 0._wp
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  va_b(ji,jj) = va_b(ji,jj) & 
                      & + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
               END DO
            END DO
            DO ji=1,jpi
               va_b(ji,jj) = va_b(ji,jj) * r1_hv_a(ji,jj)
            END DO
         END DO
      ENDIF
      !
      DO jj = mj0(jbdy1), mj1(jbdy2)
         zvb(:,jj) = 0._wp    ! Correct transport
         DO jk=1,jpkm1
            DO ji=1,jpi
               zvb(ji,jj) = zvb(ji,jj) & 
                  & + e3v_a(ji,jj,jk) * va(ji,jj,jk) * vmask(ji,jj,jk)
            END DO
         END DO
         DO ji = 1, jpi
            zvb(ji,jj) = zvb(ji,jj) * r1_hv_a(ji,jj)
         END DO

         DO jk = 1, jpkm1
            DO ji = 1, jpi
               va(ji,jj,jk) = ( va(ji,jj,jk) & 
                 & + va_b(ji,jj) - zvb(ji,jj) ) * vmask(ji,jj,jk)
            END DO
         END DO
      END DO
            
      IF( ln_dynspg_ts ) THEN       ! Set tangential velocities to time splitting estimate
         jbdy1 = jbdy1 + 1
         jbdy2 = jbdy2 + 1 
         DO jj = mj0(jbdy1), mj1(jbdy2)
            zub(:,jj) = 0._wp
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  zub(ji,jj) = zub(ji,jj) & 
                     & + e3u_a(ji,jj,jk) * ua(ji,jj,jk) * umask(ji,jj,jk)
               END DO
            END DO
            DO ji = 1, jpi
               zub(ji,jj) = zub(ji,jj) * r1_hu_a(ji,jj)
            END DO
               
            DO jk = 1, jpkm1
               DO ji = 1, jpi
                  ua(ji,jj,jk) = ( ua(ji,jj,jk) & 
                    & + ua_b(ji,jj) - zub(ji,jj) ) * umask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF
      !
   END SUBROUTINE Agrif_dyn


   SUBROUTINE Agrif_dyn_ts( jn )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_dyn_ts  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   jn
      !!
      INTEGER :: ji, jj
      INTEGER :: istart, iend, jstart, jend
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      !--- West ---!
      istart = 2
      iend   = nbghostcells+1
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            va_e(ji,jj) = vbdy(ji,jj) * hvr_e(ji,jj)
            ua_e(ji,jj) = ubdy(ji,jj) * hur_e(ji,jj)
         END DO
      END DO
      !
      !--- East ---!
      istart = jpiglo-nbghostcells
      iend   = jpiglo-1
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            va_e(ji,jj) = vbdy(ji,jj) * hvr_e(ji,jj)
         END DO
      END DO
      istart = jpiglo-nbghostcells-1
      iend   = jpiglo-2
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            ua_e(ji,jj) = ubdy(ji,jj) * hur_e(ji,jj)
         END DO
      END DO
      !
      !--- South ---!
      jstart = 2
      jend   = nbghostcells+1
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            ua_e(ji,jj) = ubdy(ji,jj) * hur_e(ji,jj)
            va_e(ji,jj) = vbdy(ji,jj) * hvr_e(ji,jj)
         END DO
      END DO
      !
      !--- North ---!
      jstart = jpjglo-nbghostcells
      jend   = jpjglo-1
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            ua_e(ji,jj) = ubdy(ji,jj) * hur_e(ji,jj)
         END DO
      END DO
      jstart = jpjglo-nbghostcells-1
      jend   = jpjglo-2
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            va_e(ji,jj) = vbdy(ji,jj) * hvr_e(ji,jj)
         END DO
      END DO
      !
   END SUBROUTINE Agrif_dyn_ts

   SUBROUTINE Agrif_dyn_ts_flux( jn, zu, zv )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_dyn_ts_flux  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   jn
      REAL(wp), DIMENSION(jpi,jpj), INTENT(inout) ::   zu, zv
      !!
      INTEGER :: ji, jj
      INTEGER :: istart, iend, jstart, jend
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      !--- West ---!
      istart = 2
      iend   = nbghostcells+1
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            zv(ji,jj) = vbdy(ji,jj) * e1v(ji,jj)
            zu(ji,jj) = ubdy(ji,jj) * e2u(ji,jj)
         END DO
      END DO
      !
      !--- East ---!
      istart = jpiglo-nbghostcells
      iend   = jpiglo-1
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            zv(ji,jj) = vbdy(ji,jj) * e1v(ji,jj)
         END DO
      END DO
      istart = jpiglo-nbghostcells-1
      iend   = jpiglo-2
      DO ji = mi0(istart), mi1(iend)
         DO jj=1,jpj
            zu(ji,jj) = ubdy(ji,jj) * e2u(ji,jj)
         END DO
      END DO
      !
      !--- South ---!
      jstart = 2
      jend   = nbghostcells+1
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            zu(ji,jj) = ubdy(ji,jj) * e2u(ji,jj)
            zv(ji,jj) = vbdy(ji,jj) * e1v(ji,jj)
         END DO
      END DO
      !
      !--- North ---!
      jstart = jpjglo-nbghostcells
      jend   = jpjglo-1
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            zu(ji,jj) = ubdy(ji,jj) * e2u(ji,jj)
         END DO
      END DO
      jstart = jpjglo-nbghostcells-1
      jend   = jpjglo-2
      DO jj = mj0(jstart), mj1(jend)
         DO ji=1,jpi
            zv(ji,jj) = vbdy(ji,jj) * e1v(ji,jj)
         END DO
      END DO
      !
   END SUBROUTINE Agrif_dyn_ts_flux

   SUBROUTINE Agrif_dta_ts( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_dta_ts  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   kt
      !!
      INTEGER :: ji, jj
      LOGICAL :: ll_int_cons
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      ll_int_cons = ln_bt_fw ! Assume conservative temporal integration in the forward case only
      !
      ! Enforce volume conservation if no time refinement:  
      IF ( Agrif_rhot()==1 ) ll_int_cons=.TRUE.  
      !
      ! Interpolate barotropic fluxes
      Agrif_SpecialValue = 0._wp
      Agrif_UseSpecialValue = ln_spc_dyn
      !
      ! Set bdy time interpolation stage to 0 (latter incremented locally do deal with corners)
      utint_stage(:,:) = 0
      vtint_stage(:,:) = 0
      !
      IF( ll_int_cons ) THEN  ! Conservative interpolation
         ! order matters here !!!!!!
         CALL Agrif_Bc_variable( ub2b_interp_id, calledweight=1._wp, procname=interpub2b ) ! Time integrated
         CALL Agrif_Bc_variable( vb2b_interp_id, calledweight=1._wp, procname=interpvb2b ) 
         !
         CALL Agrif_Bc_variable( unb_id        , calledweight=1._wp, procname=interpunb  ) ! After
         CALL Agrif_Bc_variable( vnb_id        , calledweight=1._wp, procname=interpvnb  )
         !
         CALL Agrif_Bc_variable( unb_id        , calledweight=0._wp, procname=interpunb  ) ! Before
         CALL Agrif_Bc_variable( vnb_id        , calledweight=0._wp, procname=interpvnb  )         
      ELSE ! Linear interpolation
         !
         ubdy(:,:) = 0._wp   ;   vbdy(:,:) = 0._wp 
         CALL Agrif_Bc_variable( unb_id, procname=interpunb )
         CALL Agrif_Bc_variable( vnb_id, procname=interpvnb )
      ENDIF
      Agrif_UseSpecialValue = .FALSE.
      ! 
   END SUBROUTINE Agrif_dta_ts


   SUBROUTINE Agrif_ssh( kt )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_ssh  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   kt
      !
      INTEGER  :: ji, jj
      INTEGER  :: istart, iend, jstart, jend
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !      
      ! Linear time interpolation of sea level
      !
      Agrif_SpecialValue    = 0._wp
      Agrif_UseSpecialValue = .TRUE.
      CALL Agrif_Bc_variable(sshn_id, procname=interpsshn )
      Agrif_UseSpecialValue = .FALSE.
      !
      ! --- West --- !
      istart = 2
      iend   = 1 + nbghostcells
      DO ji = mi0(istart), mi1(iend)
         DO jj = 1, jpj
            ssha(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- East --- !
      istart = jpiglo - nbghostcells
      iend   = jpiglo - 1
      DO ji = mi0(istart), mi1(iend)
         DO jj = 1, jpj
            ssha(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- South --- !
      jstart = 2
      jend   = 1 + nbghostcells
      DO jj = mj0(jstart), mj1(jend)
         DO ji = 1, jpi
            ssha(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- North --- !
      jstart = jpjglo - nbghostcells
      jend   = jpjglo - 1
      DO jj = mj0(jstart), mj1(jend)
         DO ji = 1, jpi
            ssha(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
   END SUBROUTINE Agrif_ssh


   SUBROUTINE Agrif_ssh_ts( jn )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_ssh_ts  ***
      !!----------------------------------------------------------------------  
      INTEGER, INTENT(in) ::   jn
      !!
      INTEGER :: ji, jj
      INTEGER  :: istart, iend, jstart, jend
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      ! --- West --- !
      istart = 2
      iend   = 1+nbghostcells
      DO ji = mi0(istart), mi1(iend)
         DO jj = 1, jpj
            ssha_e(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- East --- !
      istart = jpiglo - nbghostcells
      iend   = jpiglo - 1
      DO ji = mi0(istart), mi1(iend)
         DO jj = 1, jpj
            ssha_e(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- South --- !
      jstart = 2
      jend   = 1+nbghostcells
      DO jj = mj0(jstart), mj1(jend)
         DO ji = 1, jpi
            ssha_e(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
      ! --- North --- !
      jstart = jpjglo - nbghostcells
      jend   = jpjglo - 1
      DO jj = mj0(jstart), mj1(jend)
         DO ji = 1, jpi
            ssha_e(ji,jj) = hbdy(ji,jj)
         ENDDO
      ENDDO
      !
   END SUBROUTINE Agrif_ssh_ts

   SUBROUTINE Agrif_avm
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE Agrif_avm  ***
      !!----------------------------------------------------------------------  
      REAL(wp) ::   zalpha
      !!----------------------------------------------------------------------  
      !
      IF( Agrif_Root() )   RETURN
      !
      zalpha = 1._wp ! JC: proper time interpolation impossible  
                     ! => use last available value from parent 
      !
      Agrif_SpecialValue    = 0.e0
      Agrif_UseSpecialValue = .TRUE.
      !
      CALL Agrif_Bc_variable( avm_id, calledweight=zalpha, procname=interpavm )       
      !
      Agrif_UseSpecialValue = .FALSE.
      !
   END SUBROUTINE Agrif_avm
   

   SUBROUTINE interptsn( ptab, i1, i2, j1, j2, k1, k2, n1, n2, before )
      !!----------------------------------------------------------------------
      !!                  *** ROUTINE interptsn ***
      !!----------------------------------------------------------------------
      REAL(wp), DIMENSION(i1:i2,j1:j2,k1:k2,n1:n2), INTENT(inout) ::   ptab
      INTEGER                                     , INTENT(in   ) ::   i1, i2, j1, j2, k1, k2, n1, n2
      LOGICAL                                     , INTENT(in   ) ::   before
      !
      INTEGER  ::   ji, jj, jk, jn  ! dummy loop indices
      INTEGER  ::   N_in, N_out
      ! vertical interpolation:
      REAL(wp), DIMENSION(i1:i2,j1:j2,1:jpk,n1:n2) :: ptab_child
      REAL(wp), DIMENSION(k1:k2,n1:n2-1) :: tabin
      REAL(wp), DIMENSION(k1:k2) :: h_in
      REAL(wp), DIMENSION(1:jpk) :: h_out

      IF( before ) THEN         
         DO jn = 1,jpts
            DO jk=k1,k2
               DO jj=j1,j2
                 DO ji=i1,i2
                       ptab(ji,jj,jk,jn) = tsn(ji,jj,jk,jn)
                 END DO
              END DO
           END DO
        END DO

# if defined key_vertical
        DO jk=k1,k2
           DO jj=j1,j2
              DO ji=i1,i2
                 ptab(ji,jj,jk,jpts+1) = tmask(ji,jj,jk) * e3t_n(ji,jj,jk) 
              END DO
           END DO
        END DO
# endif
      ELSE 

# if defined key_vertical              
         DO jj=j1,j2
            DO ji=i1,i2
               N_in = 0
               DO jk=k1,k2 !k2 = jpk of parent grid
                  IF (ptab(ji,jj,jk,n2) == 0) EXIT
                  N_in = N_in + 1
                  tabin(jk,:) = ptab(ji,jj,jk,n1:n2-1)
                  h_in(N_in) = ptab(ji,jj,jk,n2)
               END DO
               N_out = 0
               DO jk=1,jpk ! jpk of child grid
                  IF (tmask(ji,jj,jk) == 0) EXIT 
                  N_out = N_out + 1
                  h_out(jk) = e3t_n(ji,jj,jk)
               ENDDO
               IF (N_in > 0) THEN
                  DO jn=1,jpts
                     call reconstructandremap(tabin(1:N_in,jn),h_in,ptab_child(ji,jj,1:N_out,jn),h_out,N_in,N_out)
                  ENDDO
               ENDIF
            ENDDO
         ENDDO
# else
         ptab_child(i1:i2,j1:j2,1:jpk,1:jpts) = ptab(i1:i2,j1:j2,1:jpk,1:jpts)
# endif
         !
         DO jn=1, jpts
            tsa(i1:i2,j1:j2,1:jpk,jn)=ptab_child(i1:i2,j1:j2,1:jpk,jn)*tmask(i1:i2,j1:j2,1:jpk) 
         END DO

      ENDIF
      !
   END SUBROUTINE interptsn

   SUBROUTINE interpsshn( ptab, i1, i2, j1, j2, before )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpsshn  ***
      !!----------------------------------------------------------------------  
      INTEGER                         , INTENT(in   ) ::   i1, i2, j1, j2
      REAL(wp), DIMENSION(i1:i2,j1:j2), INTENT(inout) ::   ptab
      LOGICAL                         , INTENT(in   ) ::   before
      !
      !!----------------------------------------------------------------------  
      !
      IF( before) THEN
         ptab(i1:i2,j1:j2) = sshn(i1:i2,j1:j2)
      ELSE
         hbdy(i1:i2,j1:j2) = ptab(i1:i2,j1:j2) * tmask(i1:i2,j1:j2,1)
      ENDIF
      !
   END SUBROUTINE interpsshn

   SUBROUTINE interpun( ptab, i1, i2, j1, j2, k1, k2, m1, m2, before, nb, ndir )
      !!----------------------------------------------------------------------
      !!                  *** ROUTINE interpun ***
      !!---------------------------------------------    
      !!
      INTEGER, INTENT(in) :: i1,i2,j1,j2,k1,k2,m1,m2
      REAL(wp), DIMENSION(i1:i2,j1:j2,k1:k2,m1:m2), INTENT(inout) :: ptab
      LOGICAL, INTENT(in) :: before
      INTEGER, INTENT(in) :: nb , ndir
      !!
      INTEGER :: ji,jj,jk
      REAL(wp) :: zrhoy
      ! vertical interpolation:
      REAL(wp), DIMENSION(k1:k2) :: tabin, h_in
      REAL(wp), DIMENSION(1:jpk) :: h_out
      INTEGER  :: N_in, N_out, iref
      REAL(wp) :: h_diff
      LOGICAL  :: western_side, eastern_side
      !!---------------------------------------------    
      !
      IF (before) THEN 
         DO jk=1,jpk
            DO jj=j1,j2
               DO ji=i1,i2
                  ptab(ji,jj,jk,1) = (e2u(ji,jj) * e3u_n(ji,jj,jk) * un(ji,jj,jk)*umask(ji,jj,jk)) 
# if defined key_vertical
                  ptab(ji,jj,jk,2) = (umask(ji,jj,jk) * e2u(ji,jj) * e3u_n(ji,jj,jk))
# endif
               END DO
            END DO
         END DO
      ELSE
         zrhoy = Agrif_rhoy()
# if defined key_vertical
! VERTICAL REFINEMENT BEGIN
         western_side  = (nb == 1).AND.(ndir == 1)
         eastern_side  = (nb == 1).AND.(ndir == 2)

         DO ji=i1,i2
            iref = ji
            IF (western_side) iref = MAX(2,ji)
            IF (eastern_side) iref = MIN(nlci-2,ji)
            DO jj=j1,j2
               N_in = 0
               DO jk=k1,k2
                  IF (ptab(ji,jj,jk,2) == 0) EXIT
                  N_in = N_in + 1
                  tabin(jk) = ptab(ji,jj,jk,1)/ptab(ji,jj,jk,2)
                  h_in(N_in) = ptab(ji,jj,jk,2)/(e2u(ji,jj)*zrhoy) 
              ENDDO
         
              IF (N_in == 0) THEN
                 ua(ji,jj,:) = 0._wp
                 CYCLE
              ENDIF
         
              N_out = 0
              DO jk=1,jpk
                 if (umask(iref,jj,jk) == 0) EXIT
                 N_out = N_out + 1
                 h_out(N_out) = e3u_a(iref,jj,jk)
              ENDDO
         
              IF (N_out == 0) THEN
                 ua(ji,jj,:) = 0._wp
                 CYCLE
              ENDIF
         
              IF (N_in * N_out > 0) THEN
                 h_diff = sum(h_out(1:N_out))-sum(h_in(1:N_in))
! Should be able to remove the next IF/ELSEIF statement once scale factors are dealt with properly
                 if (h_diff < -1.e4) then
                    print *,'CHECK YOUR BATHY ...', h_diff, sum(h_out(1:N_out)), sum(h_in(1:N_in))
!                    stop
                 endif
              ENDIF
              call reconstructandremap(tabin(1:N_in),h_in(1:N_in),ua(ji,jj,1:N_out),h_out(1:N_out),N_in,N_out)
            ENDDO
         ENDDO

# else
         DO jk = 1, jpkm1
            DO jj=j1,j2
               ua(i1:i2,jj,jk) = ptab(i1:i2,jj,jk,1) / ( zrhoy * e2u(i1:i2,jj) * e3u_a(i1:i2,jj,jk) )
            END DO
         END DO
# endif

      ENDIF
      ! 
   END SUBROUTINE interpun

   SUBROUTINE interpvn( ptab, i1, i2, j1, j2, k1, k2, m1, m2, before, nb, ndir )
      !!----------------------------------------------------------------------
      !!                  *** ROUTINE interpvn ***
      !!----------------------------------------------------------------------
      !
      INTEGER, INTENT(in) :: i1,i2,j1,j2,k1,k2,m1,m2
      REAL(wp), DIMENSION(i1:i2,j1:j2,k1:k2,m1:m2), INTENT(inout) :: ptab
      LOGICAL, INTENT(in) :: before
      INTEGER, INTENT(in) :: nb , ndir
      !
      INTEGER :: ji,jj,jk
      REAL(wp) :: zrhox
      ! vertical interpolation:
      REAL(wp), DIMENSION(k1:k2) :: tabin, h_in
      REAL(wp), DIMENSION(1:jpk) :: h_out
      INTEGER  :: N_in, N_out, jref
      REAL(wp) :: h_diff
      LOGICAL  :: northern_side,southern_side
      !!---------------------------------------------    
      !      
      IF (before) THEN          
         DO jk=k1,k2
            DO jj=j1,j2
               DO ji=i1,i2
                  ptab(ji,jj,jk,1) = (e1v(ji,jj) * e3v_n(ji,jj,jk) * vn(ji,jj,jk)*vmask(ji,jj,jk))
# if defined key_vertical
                  ptab(ji,jj,jk,2) = vmask(ji,jj,jk) * e1v(ji,jj) * e3v_n(ji,jj,jk)
# endif
               END DO
            END DO
         END DO
      ELSE       
         zrhox = Agrif_rhox()
# if defined key_vertical

         southern_side = (nb == 2).AND.(ndir == 1)
         northern_side = (nb == 2).AND.(ndir == 2)

         DO jj=j1,j2
            jref = jj
            IF (southern_side) jref = MAX(2,jj)
            IF (northern_side) jref = MIN(nlcj-2,jj)
            DO ji=i1,i2
               N_in = 0
               DO jk=k1,k2
                  if (ptab(ji,jj,jk,2) == 0) EXIT
                  N_in = N_in + 1
                  tabin(jk) = ptab(ji,jj,jk,1)/ptab(ji,jj,jk,2)
                  h_in(N_in) = ptab(ji,jj,jk,2)/(e1v(ji,jj)*zrhox)
               END DO
               IF (N_in == 0) THEN
                  va(ji,jj,:) = 0._wp
                  CYCLE
               ENDIF
         
               N_out = 0
               DO jk=1,jpk
                  if (vmask(ji,jref,jk) == 0) EXIT
                  N_out = N_out + 1
                  h_out(N_out) = e3v_a(ji,jref,jk)
               END DO
               IF (N_out == 0) THEN
                 va(ji,jj,:) = 0._wp
                 CYCLE
               ENDIF
               call reconstructandremap(tabin(1:N_in),h_in(1:N_in),va(ji,jj,1:N_out),h_out(1:N_out),N_in,N_out)
            END DO
         END DO
# else
         DO jk = 1, jpkm1
            va(i1:i2,j1:j2,jk) = ptab(i1:i2,j1:j2,jk,1) / ( zrhox * e1v(i1:i2,j1:j2) * e3v_a(i1:i2,j1:j2,jk) )
         END DO
# endif
      ENDIF
      !        
   END SUBROUTINE interpvn

   SUBROUTINE interpunb( ptab, i1, i2, j1, j2, before)
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpunb  ***
      !!----------------------------------------------------------------------  
      INTEGER                         , INTENT(in   ) ::   i1, i2, j1, j2
      REAL(wp), DIMENSION(i1:i2,j1:j2), INTENT(inout) ::   ptab
      LOGICAL                         , INTENT(in   ) ::   before
      !
      INTEGER  ::   ji, jj
      REAL(wp) ::   zrhoy, zrhot, zt0, zt1, ztcoeff
      !!----------------------------------------------------------------------  
      !
      IF( before ) THEN 
         ptab(i1:i2,j1:j2) = e2u(i1:i2,j1:j2) * hu_n(i1:i2,j1:j2) * un_b(i1:i2,j1:j2)
      ELSE
         zrhoy = Agrif_Rhoy()
         zrhot = Agrif_rhot()
         ! Time indexes bounds for integration
         zt0 = REAL(Agrif_NbStepint()  , wp) / zrhot
         zt1 = REAL(Agrif_NbStepint()+1, wp) / zrhot      
         ! 
         DO ji = i1, i2
            DO jj = j1, j2
               IF    ( utint_stage(ji,jj) == 1  ) THEN
                  ztcoeff = zrhot * (  zt1**2._wp * (       zt1 - 1._wp)        &
                     &               - zt0**2._wp * (       zt0 - 1._wp)        )
               ELSEIF( utint_stage(ji,jj) == 2  ) THEN
                  ztcoeff = zrhot * (  zt1        * (       zt1 - 1._wp)**2._wp &
                     &               - zt0        * (       zt0 - 1._wp)**2._wp )
               ELSEIF( utint_stage(ji,jj) == 0  ) THEN                
                  ztcoeff = 1._wp
               ELSE
                  ztcoeff = 0._wp
               ENDIF
               !   
               ubdy(ji,jj) = ubdy(ji,jj) + ztcoeff * ptab(ji,jj)
               !            
               IF (( utint_stage(ji,jj) == 2 ).OR.( utint_stage(ji,jj) == 0 )) THEN
                  ubdy(ji,jj) = ubdy(ji,jj) / (zrhoy*e2u(ji,jj)) * umask(ji,jj,1)
                  utint_stage(ji,jj) = 3
               ELSE
                  utint_stage(ji,jj) = utint_stage(ji,jj) + 1
               ENDIF
            END DO
         END DO
      END IF
      ! 
   END SUBROUTINE interpunb


   SUBROUTINE interpvnb( ptab, i1, i2, j1, j2, before )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpvnb  ***
      !!----------------------------------------------------------------------  
      INTEGER                         , INTENT(in   ) ::   i1, i2, j1, j2
      REAL(wp), DIMENSION(i1:i2,j1:j2), INTENT(inout) ::   ptab
      LOGICAL                         , INTENT(in   ) ::   before
      !
      INTEGER  ::   ji,jj
      REAL(wp) ::   zrhox, zrhot, zt0, zt1, ztcoeff   
      !!----------------------------------------------------------------------  
      ! 
      IF( before ) THEN 
         ptab(i1:i2,j1:j2) = e1v(i1:i2,j1:j2) * hv_n(i1:i2,j1:j2) * vn_b(i1:i2,j1:j2)
      ELSE
         zrhox = Agrif_Rhox()
         zrhot = Agrif_rhot()
         ! Time indexes bounds for integration
         zt0 = REAL(Agrif_NbStepint()  , wp) / zrhot
         zt1 = REAL(Agrif_NbStepint()+1, wp) / zrhot 
         !     
         DO ji = i1, i2
            DO jj = j1, j2
               IF    ( vtint_stage(ji,jj) == 1  ) THEN
                  ztcoeff = zrhot * (  zt1**2._wp * (       zt1 - 1._wp)        &
                     &               - zt0**2._wp * (       zt0 - 1._wp)        )
               ELSEIF( vtint_stage(ji,jj) == 2  ) THEN
                  ztcoeff = zrhot * (  zt1        * (       zt1 - 1._wp)**2._wp &
                     &               - zt0        * (       zt0 - 1._wp)**2._wp )
               ELSEIF( vtint_stage(ji,jj) == 0  ) THEN                
                  ztcoeff = 1._wp
               ELSE
                  ztcoeff = 0._wp
               ENDIF
               !   
               vbdy(ji,jj) = vbdy(ji,jj) + ztcoeff * ptab(ji,jj)
               !            
               IF (( vtint_stage(ji,jj) == 2 ).OR.( vtint_stage(ji,jj) == 0 )) THEN
                  vbdy(ji,jj) = vbdy(ji,jj) / (zrhox*e1v(ji,jj)) * vmask(ji,jj,1)
                  vtint_stage(ji,jj) = 3
               ELSE
                  vtint_stage(ji,jj) = vtint_stage(ji,jj) + 1
               ENDIF
            END DO
         END DO          
      ENDIF
      !
   END SUBROUTINE interpvnb


   SUBROUTINE interpub2b( ptab, i1, i2, j1, j2, before )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpub2b  ***
      !!----------------------------------------------------------------------  
      INTEGER                         , INTENT(in   ) ::   i1, i2, j1, j2
      REAL(wp), DIMENSION(i1:i2,j1:j2), INTENT(inout) ::   ptab
      LOGICAL                         , INTENT(in   ) ::   before
      !
      INTEGER  ::   ji,jj
      REAL(wp) ::   zrhot, zt0, zt1, zat
      !!----------------------------------------------------------------------  
      IF( before ) THEN
         IF ( ln_bt_fw ) THEN
            ptab(i1:i2,j1:j2) = e2u(i1:i2,j1:j2) * ub2_b(i1:i2,j1:j2)
         ELSE
            ptab(i1:i2,j1:j2) = e2u(i1:i2,j1:j2) * un_adv(i1:i2,j1:j2)
         ENDIF
      ELSE
         zrhot = Agrif_rhot()
         ! Time indexes bounds for integration
         zt0 = REAL(Agrif_NbStepint()  , wp) / zrhot
         zt1 = REAL(Agrif_NbStepint()+1, wp) / zrhot
         ! Polynomial interpolation coefficients:
         zat = zrhot * (  zt1**2._wp * (-2._wp*zt1 + 3._wp)    &
            &           - zt0**2._wp * (-2._wp*zt0 + 3._wp)    ) 
         !
         ubdy(i1:i2,j1:j2) = zat * ptab(i1:i2,j1:j2) 
         !
         ! Update interpolation stage:
         utint_stage(i1:i2,j1:j2) = 1
      ENDIF
      ! 
   END SUBROUTINE interpub2b
   

   SUBROUTINE interpvb2b( ptab, i1, i2, j1, j2, before )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpvb2b  ***
      !!----------------------------------------------------------------------  
      INTEGER                         , INTENT(in   ) ::   i1, i2, j1, j2
      REAL(wp), DIMENSION(i1:i2,j1:j2), INTENT(inout) ::   ptab
      LOGICAL                         , INTENT(in   ) ::   before
      !
      INTEGER ::   ji,jj
      REAL(wp) ::   zrhot, zt0, zt1, zat
      !!----------------------------------------------------------------------  
      !
      IF( before ) THEN
         IF ( ln_bt_fw ) THEN
            ptab(i1:i2,j1:j2) = e1v(i1:i2,j1:j2) * vb2_b(i1:i2,j1:j2)
         ELSE
            ptab(i1:i2,j1:j2) = e1v(i1:i2,j1:j2) * vn_adv(i1:i2,j1:j2)
         ENDIF
      ELSE      
         zrhot = Agrif_rhot()
         ! Time indexes bounds for integration
         zt0 = REAL(Agrif_NbStepint()  , wp) / zrhot
         zt1 = REAL(Agrif_NbStepint()+1, wp) / zrhot
         ! Polynomial interpolation coefficients:
         zat = zrhot * (  zt1**2._wp * (-2._wp*zt1 + 3._wp)    &
            &           - zt0**2._wp * (-2._wp*zt0 + 3._wp)    ) 
         !
         vbdy(i1:i2,j1:j2) = zat * ptab(i1:i2,j1:j2)
         !
         ! update interpolation stage:
         vtint_stage(i1:i2,j1:j2) = 1
      ENDIF
      !      
   END SUBROUTINE interpvb2b


   SUBROUTINE interpe3t( ptab, i1, i2, j1, j2, k1, k2, before, nb, ndir )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpe3t  ***
      !!----------------------------------------------------------------------  
      INTEGER                              , INTENT(in   ) :: i1, i2, j1, j2, k1, k2
      REAL(wp),DIMENSION(i1:i2,j1:j2,k1:k2), INTENT(inout) :: ptab
      LOGICAL                              , INTENT(in   ) :: before
      INTEGER                              , INTENT(in   ) :: nb , ndir
      !
      INTEGER :: ji, jj, jk
      LOGICAL :: western_side, eastern_side, northern_side, southern_side
      !!----------------------------------------------------------------------  
      !    
      IF( before ) THEN
         ptab(i1:i2,j1:j2,k1:k2) = tmask(i1:i2,j1:j2,k1:k2) * e3t_0(i1:i2,j1:j2,k1:k2)
      ELSE
         western_side  = (nb == 1).AND.(ndir == 1)
         eastern_side  = (nb == 1).AND.(ndir == 2)
         southern_side = (nb == 2).AND.(ndir == 1)
         northern_side = (nb == 2).AND.(ndir == 2)
         !
         DO jk = k1, k2
            DO jj = j1, j2
               DO ji = i1, i2
                  !
                  IF( ABS( ptab(ji,jj,jk) - tmask(ji,jj,jk) * e3t_0(ji,jj,jk) ) > 1.D-2) THEN
                     IF (western_side.AND.(ptab(i1+nbghostcells-1,jj,jk)>0._wp)) THEN
                        WRITE(numout,*) 'ERROR bathymetry merge at the western border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*)  ptab(ji,jj,jk), e3t_0(ji,jj,jk) 
                        kindic_agr = kindic_agr + 1
                     ELSEIF (eastern_side.AND.(ptab(i2-nbghostcells+1,jj,jk)>0._wp)) THEN
                        WRITE(numout,*) 'ERROR bathymetry merge at the eastern border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*)  ptab(ji,jj,jk), e3t_0(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ELSEIF (southern_side.AND.(ptab(ji,j1+nbghostcells-1,jk)>0._wp)) THEN
                        WRITE(numout,*) 'ERROR bathymetry merge at the southern border ji,jj,jk', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*)  ptab(ji,jj,jk), e3t_0(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ELSEIF (northern_side.AND.(ptab(ji,j2-nbghostcells+1,jk)>0._wp)) THEN
                        WRITE(numout,*) 'ERROR bathymetry merge at the northen border ji,jj,jk', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*)  ptab(ji,jj,jk), e3t_0(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         !
      ENDIF
      ! 
   END SUBROUTINE interpe3t


   SUBROUTINE interpumsk( ptab, i1, i2, j1, j2, k1, k2, before, nb, ndir )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpumsk  ***
      !!----------------------------------------------------------------------  
      INTEGER                              , INTENT(in   ) ::   i1, i2, j1, j2, k1, k2
      REAL(wp),DIMENSION(i1:i2,j1:j2,k1:k2), INTENT(inout) ::   ptab
      LOGICAL                              , INTENT(in   ) ::   before
      INTEGER                              , INTENT(in   ) ::   nb , ndir
      !
      INTEGER ::   ji, jj, jk
      LOGICAL ::   western_side, eastern_side   
      !!----------------------------------------------------------------------  
      !    
      IF( before ) THEN
         ptab(i1:i2,j1:j2,k1:k2) = umask(i1:i2,j1:j2,k1:k2)
      ELSE
         western_side = (nb == 1).AND.(ndir == 1)
         eastern_side = (nb == 1).AND.(ndir == 2)
         DO jk = k1, k2
            DO jj = j1, j2
               DO ji = i1, i2
                   ! Velocity mask at boundary edge points:
                  IF (ABS(ptab(ji,jj,jk) - umask(ji,jj,jk)) > 1.D-2) THEN
                     IF (western_side) THEN
                        WRITE(numout,*) 'ERROR with umask at the western border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*) '      masks: parent, child ', ptab(ji,jj,jk), umask(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ELSEIF (eastern_side) THEN
                        WRITE(numout,*) 'ERROR with umask at the eastern border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*) '      masks: parent, child ', ptab(ji,jj,jk), umask(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         !
      ENDIF
      ! 
   END SUBROUTINE interpumsk


   SUBROUTINE interpvmsk( ptab, i1, i2, j1, j2, k1, k2, before, nb, ndir )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interpvmsk  ***
      !!----------------------------------------------------------------------  
      INTEGER                              , INTENT(in   ) ::   i1,i2,j1,j2,k1,k2
      REAL(wp),DIMENSION(i1:i2,j1:j2,k1:k2), INTENT(inout) ::   ptab
      LOGICAL                              , INTENT(in   ) ::   before
      INTEGER                              , INTENT(in   ) :: nb , ndir
      !
      INTEGER ::   ji, jj, jk
      LOGICAL ::   northern_side, southern_side     
      !!----------------------------------------------------------------------  
      !    
      IF( before ) THEN
         ptab(i1:i2,j1:j2,k1:k2) = vmask(i1:i2,j1:j2,k1:k2)
      ELSE
         southern_side = (nb == 2).AND.(ndir == 1)
         northern_side = (nb == 2).AND.(ndir == 2)
         DO jk = k1, k2
            DO jj = j1, j2
               DO ji = i1, i2
                   ! Velocity mask at boundary edge points:
                  IF (ABS(ptab(ji,jj,jk) - vmask(ji,jj,jk)) > 1.D-2) THEN
                     IF (southern_side) THEN
                        WRITE(numout,*) 'ERROR with vmask at the southern border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*) '      masks: parent, child ', ptab(ji,jj,jk), vmask(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ELSEIF (northern_side) THEN
                        WRITE(numout,*) 'ERROR with vmask at the northern border ji,jj,jk ', ji+nimpp-1,jj+njmpp-1,jk
                        WRITE(numout,*) '      masks: parent, child ', ptab(ji,jj,jk), vmask(ji,jj,jk)
                        kindic_agr = kindic_agr + 1
                     ENDIF
                  ENDIF
               END DO
            END DO
         END DO
         !
      ENDIF
      ! 
   END SUBROUTINE interpvmsk


   SUBROUTINE interpavm( ptab, i1, i2, j1, j2, k1, k2, m1, m2, before )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE interavm  ***
      !!----------------------------------------------------------------------  
      INTEGER                                    , INTENT(in   ) ::   i1, i2, j1, j2, k1, k2, m1, m2
      REAL(wp),DIMENSION(i1:i2,j1:j2,k1:k2,m1:m2), INTENT(inout) ::   ptab
      LOGICAL                                    , INTENT(in   ) ::   before
      REAL(wp), DIMENSION(k1:k2) :: tabin, h_in
      REAL(wp), DIMENSION(1:jpk) :: h_out
      INTEGER  :: N_in, N_out, ji, jj, jk
      !!----------------------------------------------------------------------  
      !      
      IF (before) THEN         
         DO jk=k1,k2
            DO jj=j1,j2
              DO ji=i1,i2
                    ptab(ji,jj,jk,1) = avm_k(ji,jj,jk)
              END DO
           END DO
        END DO
#ifdef key_vertical         
        DO jk=k1,k2
           DO jj=j1,j2
              DO ji=i1,i2
                 ptab(ji,jj,jk,2) = wmask(ji,jj,jk) * e3w_n(ji,jj,jk) 
              END DO
           END DO
        END DO
#endif
      ELSE 
#ifdef key_vertical         
         avm_k(i1:i2,j1:j2,1:jpk) = 0.
         DO jj=j1,j2
            DO ji=i1,i2
               N_in = 0
               DO jk=k1,k2 !k2 = jpk of parent grid
                  IF (ptab(ji,jj,jk,2) == 0) EXIT
                  N_in = N_in + 1
                  tabin(jk) = ptab(ji,jj,jk,1)
                  h_in(N_in) = ptab(ji,jj,jk,2)
               END DO
               N_out = 0
               DO jk=1,jpk ! jpk of child grid
                  IF (wmask(ji,jj,jk) == 0) EXIT 
                  N_out = N_out + 1
                  h_out(jk) = e3t_n(ji,jj,jk)
               ENDDO
               IF (N_in > 0) THEN
                  CALL reconstructandremap(tabin(1:N_in),h_in,avm_k(ji,jj,1:N_out),h_out,N_in,N_out)
               ENDIF
            ENDDO
         ENDDO
#else
         avm_k(i1:i2,j1:j2,k1:k2) = ptab (i1:i2,j1:j2,k1:k2,1)
#endif
      ENDIF
      !
   END SUBROUTINE interpavm

#else
   !!----------------------------------------------------------------------
   !!   Empty module                                          no AGRIF zoom
   !!----------------------------------------------------------------------
CONTAINS
   SUBROUTINE Agrif_OCE_Interp_empty
      WRITE(*,*)  'agrif_oce_interp : You should not have seen this print! error?'
   END SUBROUTINE Agrif_OCE_Interp_empty
#endif

   !!======================================================================
END MODULE agrif_oce_interp
